# Copyright 2004-2009 Canonical Ltd.  All rights reserved.
#
# This file is part of lazr.batchnavigator
#
# lazr.batchnavigator is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# lazr.batchnavigator is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with lazr.batchnavigator. If not, see <http://www.gnu.org/licenses/>.

"""Functions for working with generic syntax URIs."""

try:
    import importlib.metadata as importlib_metadata
except ImportError:
    import importlib_metadata

__version__ = importlib_metadata.version('lazr.batchnavigator')

# While we generally frown on "*" imports, this, combined with the fact we
# only test code from this module, means that we can verify what has been
# exported.
from lazr.batchnavigator._batchnavigator import *
from lazr.batchnavigator._batchnavigator import __all__



# Copyright 2004-2010 Canonical Ltd.  All rights reserved.
#
# This file is part of lazr.batchnavigator
#
# lazr.batchnavigator is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# lazr.batchnavigator is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with lazr.batchnavigator. If not, see <http://www.gnu.org/licenses/>.

__metaclass__ = type

from six.moves.urllib.parse import (
    parse_qs,
    urlencode,
    )
from zope.interface import implementer, provider
from zope.interface.common.sequence import IFiniteSequence
from zope.cachedescriptors.property import Lazy

from lazr.batchnavigator.z3batching.batch import _Batch
from lazr.batchnavigator.interfaces import (
    IBatchNavigator,
    IBatchNavigatorFactory,
    IRangeFactory,
    InvalidBatchSizeError,
    )

__all__ = [
    'BatchNavigator',
    'ListRangeFactory',
    ]

@implementer(IBatchNavigator)
@provider(IBatchNavigatorFactory)
class BatchNavigator:

    # subclasses can override
    _batch_factory = _Batch

    start_variable_name = 'start'
    batch_variable_name = 'batch'
    memo_variable_name = 'memo'
    direction_variable_name = 'direction'
    # Set to e.g. 'active' to make the variable names become 'active_start',
    # 'active_batch' etc.
    variable_name_prefix = ''

    # The size the batch navigator was constructed with
    default_size = None
    # The size used if no specific size was supplied to the constructor.
    default_batch_size = 50
    max_batch_size = 300

    # We want subclasses to be able to hide the 'Last' link from
    # users.  They may want to do this for really large result sets;
    # for example, batches with over a hundred thousand items.
    show_last_link = True

    # The default heading describing the kind of objects in the batch.
    # Sub-classes can override this to be more specific.
    default_singular_heading = 'result'
    default_plural_heading = 'results'

    transient_parameters = None

    @Lazy
    def query_string_parameters(self):
        query_string = self.request.get('QUERY_STRING', '')

        # Just in case QUERY_STRING is in the environment explicitly as
        # None (Some tests seem to do this, but not sure if it can ever
        # happen outside of tests.)
        if query_string is None:
            query_string = ''
        return parse_qs(query_string, keep_blank_values=True)

    def __init__(self, results, request, start=0, size=None, callback=None,
                 transient_parameters=None, force_start=False,
                 range_factory=None):
        "See `IBatchNavigatorFactory.__call__`"
        self.request = request
        self._update_variable_names()
        local = (self.batch_variable_name, self.start_variable_name,
            self.memo_variable_name, self.direction_variable_name)
        self.transient_parameters = set(local)
        if transient_parameters is not None:
            self.transient_parameters.update(transient_parameters)

        # For backwards compatibility (as in the past a work-around has been
        # to include the url batch params in hidden fields within posted
        # forms), if the request is a POST request, and either the 'start'
        # or 'batch' params are included then revert to the default behaviour
        # of using the request (which automatically gets the params from the
        # request.form dict).
        if request.method == 'POST' and (
            self.start_variable_name in request.form or
            self.batch_variable_name in request.form):
            batch_params_source = request
        else:
            # We grab the request variables directly from the requests
            # query_string_parameters so that they will be recognized
            # even during post operations.
            batch_params_source = dict(
                (k, v[0]) for k, v
                in self.query_string_parameters.items() if k in local)

        # In this code we ignore invalid request variables since it
        # probably means the user finger-fumbled it in the request. We
        # could raise UnexpectedFormData, but is there a good reason?
        def param_var(name):
            return batch_params_source.get(name, None)
        # -- start
        request_start = param_var(self.start_variable_name)
        if force_start or request_start is None:
            self.start = start
        else:
            try:
                self.start = int(request_start)
            except (ValueError, TypeError):
                self.start = start
        # -- size
        size = self.determineSize(size, batch_params_source)
        # -- direction
        direction = param_var(self.direction_variable_name)
        if direction == 'backwards':
            direction = False
        else:
            direction = None
        # -- memo
        memo = param_var(self.memo_variable_name)
        if direction is not None and memo is None:
            # Walking backwards from the end - the only case where we generate
            # a url with no memo but a direction (and the only case where we
            # need it: from the start with no memo is equivalent to a simple
            # list slice anyway).
            memo = ''

        if range_factory is None:
            range_factory = ListRangeFactory(results)
        self.batch = self._batch_factory(results, range_factory,
            start=self.start, size=size, range_forwards=direction,
            range_memo=memo)
        if callback is not None:
            callback(self, self.batch)
        self.setHeadings(
            self.default_singular_heading, self.default_plural_heading)

    def determineSize(self, size, batch_params_source):
        """Determine the default and user requested batch sizes.

        This function should assign the default size for the batch to
        self.default_size. The base class implementation uses the size passed
        to the constructor, but other implementations may choose to clamp it or
        force a particular default size.

        :param size: Size passed to the constructor.
        :param batch_params_source: User parameters dict.
        :return: The size to be used for this batch.
        """
        self.default_size = size
        request_size = self._getRequestedSize(batch_params_source)
        if request_size is not None:
            size = request_size
        if size is None:
            size = self.default_batch_size
        return size

    def _getRequestedSize(self, batch_params_source):
        """Figure out what batch size the user requested, if any.

        Sizes that are not positive numbers are ignored.

        :return: An acceptable batch size requested by the user, or
            None.
        :raise: `InvalidBatchSizeError` if the requested size exceeds
            `max_batch_size`.
        """
        size_string = batch_params_source.get(self.batch_variable_name, None)
        if size_string is None:
            return None

        try:
            request_size = int(size_string)
        except (ValueError, TypeError):
            return None

        if request_size <= 0:
            return None

        if request_size > self.max_batch_size:
            raise InvalidBatchSizeError(
                'Maximum for "%s" parameter is %d.' %
                (self.batch_variable_name,
                 self.max_batch_size))

        return request_size

    @property
    def heading(self):
        """See `IBatchNavigator`"""
        if self.batch.total() == 1:
            return self._singular_heading
        return self._plural_heading

    def setHeadings(self, singular, plural):
        """See `IBatchNavigator`"""
        self._singular_heading = singular
        self._plural_heading = plural

    def getCleanQueryParams(self, params=None):
        """Removes batch nav params if present and returns a sequence
        of key-values pairs.

        If ``params`` is None, uses the current query_string_params.
        """
        if params is None:
            params = []
            for k, v in self.query_string_parameters.items():
                params.extend((k, item) for item in v)
        else:
            try:
                params = params.items()
            except AttributeError:
                pass

        # We need the doseq=True because some url params are for multi-value
        # fields.
        return [
            (key, value) for (key, value) in sorted(params)
            if key not in self.transient_parameters]

    def getCleanQueryString(self, params=None):
        """Removes batch nav params if present and returns a query
        string.

        If ``params`` is None, uses the current query_string_params.
        """
        # We need the doseq=True because some url params are for multi-value
        # fields.
        return urlencode(self.getCleanQueryParams(params), doseq=True)

    def generateBatchURL(self, batch, backwards=False):
        url = ""
        if batch is None:
            return url

        params = self.getCleanQueryParams()

        size = batch.size
        if size != self.default_size:
            # The current batch size should only be part of the URL if it's
            # different from the default batch size.
            params.append((self.batch_variable_name, size))

        if backwards:
            params.append((self.direction_variable_name, "backwards"))

        if batch.range_memo:
            params.append((self.memo_variable_name, batch.range_memo))

        start = batch.startNumber() - 1
        if start:
            params.append((self.start_variable_name, start))

        base_url = str(self.request.URL)
        return "%s?%s" % (base_url, urlencode(params))

    def firstBatchURL(self):
        batch = self.batch.firstBatch()
        if not self.batch.has_previous_batch:
            # We are already on the first batch.
            batch = None
        return self.generateBatchURL(batch)

    def prevBatchURL(self):
        return self.generateBatchURL(self.batch.prevBatch(), backwards=True)

    def nextBatchURL(self):
        return self.generateBatchURL(self.batch.nextBatch())

    def lastBatchURL(self):
        batch = self.batch.lastBatch()
        if not self.batch.has_next_batch:
            # We are already on the last batch.
            batch = None
        return self.generateBatchURL(batch, backwards=True)

    def currentBatch(self):
        return self.batch

    def _update_variable_names(self):
        """Update self.x_variable_name with self.variable_name_prefix.

        This gives the concrete instance the same prefix for all variables.
        """
        prefix = self.variable_name_prefix or ''
        if prefix:
            prefix += '_'
        for varname in ('start', 'batch', 'memo', 'direction'):
            attrname = varname + '_variable_name'
            setattr(self, attrname, prefix + getattr(self, attrname))


@implementer(IRangeFactory)
class ListRangeFactory:
    """Implements an IRangeFactory for lists (and list-like objects).

    This uses the slice protocol index as its memos: 'up to and not including
    low', 'from this point and up'.
    """

    def __init__(self, results):
        self.results = results

    def getEndpointMemos(self, batch):
        """See `IRangeFactory`

        Most implementations will want to use batch.sliced_list to retrieve
        database keys.
        """
        end_idx = batch.trueSize + batch.start
        return (str(batch.start), str(end_idx))

    def getSlice(self, size, endpoint_memo=None, forwards=True):
        """See `IRangeFactory`"""
        if not size:
            return []
        if not forwards:
            size = -size
        if endpoint_memo:
            try:
                offset = int(endpoint_memo)
            except ValueError:
                raise InvalidBatchSizeError('not an int')
            offset_plus_size = offset + size
            if offset_plus_size < 1:
                offset_plus_size = 0
        else:
            offset = None
            offset_plus_size = size
        if forwards:
            return list(self.results[offset:offset_plus_size])
        else:
            if offset is None:
                # SQL mapped result sets will blow up on [-N:None] slices, so
                # get the total size and create absolute references.
                # This is only used for the 'Last' link in a collection, and
                # any nontrivial collection shouldn't be using ListRangeFactory
                # unless indexing is cheap (which implies counting is cheap).
                # If this becomes an issue, a shared length cache can be built
                # with _Batch, but because other implementations of
                # IRangeFactory would need to honour that protocol, it should
                # be avoided unless needed.
                if getattr(self.results, '__len__', None) is None:
                    self.results = IFiniteSequence(self.results)
                total_length = len(self.results)
                offset = total_length
                # Storm raises an exception for a negative initial offset.
                offset_plus_size = max(offset + offset_plus_size, 0)
            if offset_plus_size == offset:
                # Save a query if the slice is empty.
                # offset cannot be None so we don't have to deal with the
                # effect of None parameters as slice boundaries.
                return []
            else:
                result = list(self.results[offset_plus_size:offset])
                result.reverse()
                return result

    def getSliceByIndex(self, start, end):
        """See `IRangeFactory`"""
        if self.results is not None:
            return list(self.results[start:end])
        else:
            return []

    @Lazy
    def rough_length(self):
        """See `IRangeFactory`"""
        results = self.results
        # LBYL: we don't want to mask exceptions that len might raise, and we
        # don't have chained exceptions yet.
        if getattr(results, '__len__', None) is None:
            results = IFiniteSequence(results)
        return len(results)



##############################################################################
#
# Copyright (c) 2003 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
# Some parts Copyright 2005 Canonical Ltd.
#
##############################################################################
"""
Zope-derived Batching Support

*Do not use this. It is now an internal implementation detail.*
The intent is to get rid of this in favor of moving to extending z3c.batching.
"""

from __future__ import absolute_import

from zope.interface import implementer
from zope.interface.common.sequence import IFiniteSequence
from .interfaces import IBatch

from zope.cachedescriptors.property import Lazy

# The base batch size, which can be overridden by users of _Batch such
# as BatchNavigator. In Launchpad, we override it via a config option.
BATCH_SIZE = 50

@implementer(IBatch)
class _PartialBatch:
    """A helper batch implementation.

    _Batch.sliced_list() below needs to retrieve a second chunk of
    data when a call of range_factory.getSlice() for a backwards batch
    returns less elements than requested because the start of the result
    set is reached.

    In this case, another (forwards) batch must be retrieved. _Batch
    must not assume that the memo value used to retrieve the first, too
    small, result set can be used to retrieve the additional data. (The
    class RangeFactoryWithValueBasedEndpointMemos in
    tests/test_z3batching.py is an example where this assumption fails.)

    Instead, _Batch.sliced_list() must retrieve the endpoint memos for
    the partial data and use them to retrieve the missing part of the
    result set. Since
      - IRangeFactory.getEndpointMemos(batch) is free to use any
        property of IBatch,
      - a call like self.range_factory.getEndpointMemos(self) in
       _Batch.sliced_list() leads to infinite recursions if the
       range factory wants to access sliced_list,

    we use this helper class for the getEndpointMemos() call in
    _Batch.sliced_list().
    """

    def __init__(self, sliced_list):
        self.start = 0
        self.trueSize = len(sliced_list)
        self.sliced_list = sliced_list
        self.size = len(sliced_list)

    def __len__(self):
        """See `IBatch`."""
        return len(self.sliced_list)

    def __iter__(self):
        """See `IBatch`."""
        return iter(self.sliced_list)

    def __getitem__(self, index):
        """See `IBatch`."""
        return self.sliced_list[index]

    def __contains__(self, key):
        """See `IBatch`."""
        return 0 <= key < len(self.sliced_list)

    def nextBatch(self):
        """See `IBatch`."""
        raise NotImplementedError

    def prevBatch(self):
        """See `IBatch`."""
        raise NotImplementedError

    def first(self):
        """See `IBatch`."""
        return self.sliced_list[0]

    def last(self):
        """See `IBatch`."""
        return self.sliced_list[-1]

    def total(self):
        """See `IBatch`."""
        return len(self.sliced_list)

    def startNumber(self):
        """See `IBatch`."""
        return 1

    def endNumber(self):
        """See `IBatch`."""
        return len(self.sliced_list) + 1

    # The values do not matter at all. Just make verifyObject() happy.
    has_previous_batch = None
    has_next_batch = None


@implementer(IBatch)
class _Batch(object):

    def _get_length(self, results):
        # If this batch's contents is smaller than the batch size, it is the
        # last batch and we can deduce the underlying data's size.
        if (self._sliced_list is not None
                and len(self._sliced_list) <= self.size):
            return self.start + len(self._sliced_list)
        return self.range_factory.rough_length

    @property
    def listlength(self):
        if self._listlength is None:
            self._listlength = self._get_length(self.list)
        return self._listlength

    @Lazy
    def trueSize(self):
        """Return the actual size of this batch."""
        length = len(self.sliced_list)
        if length >= self.size:
            # This batch is full (and there might be another batch
            # afterwards). Return .size
            return self.size
        else:
            # This batch is not full. Return its (user-visible)
            # length.
            return length

    def __init__(self, results, range_factory, start=0, size=None,
        range_forwards=None, range_memo=None, _listlength=None):
        """Create a _Batch.

        :param start: Cosmetic indicator of the start of the batch. Has *no*
            effect on returned data.
        :param range_factory: A factory used to construct efficient views of
            the results.
        :param range_forwards: True if the range memo is at or before the start
            of the batch.
        :param range_memo: An endpoint memo from the range factory describing
            how to get a slice from the factory which will have offset 0
            contain the first element of the batch (or the list, if not
            range_forwards). '' for the extreme edge of the range, None to
            have the start parameter take precedence and cause actual
            slicing of results.
        """
        if results is None:
            results = []
        self.list = results

        if _listlength is not None:
            self._listlength = _listlength
        else:
            self._listlength = None
        self._sliced_list = None

        if size is None:
            size = BATCH_SIZE
        self.size = size

        self.start = start
        self.end = start + size
        self.range_factory = range_factory
        self.range_memo = range_memo
        if range_forwards is None:
            range_forwards = True
        self.range_forwards = range_forwards
        self.is_first_batch = False

    def __len__(self):
        assert self.trueSize >= 0, ('The number of items in a batch should '
            'never be negative.')
        return self.trueSize

    def __getitem__(self, key):
        if key >= self.trueSize:
            raise IndexError('batch index out of range')
        # When self.start is negative (IOW, when we are batching over an
        # empty list) we need to raise IndexError here; otherwise, the
        # attempt to slice below will, on objects which don't implement
        # __len__, raise a mysterious exception directly from python.
        if self.start < 0:
            raise IndexError('batch is empty')
        # We delegate to self.sliced_list because:
        # - usually folk will be walking the batch and sliced_list optimises
        #   that.
        # - if they aren't, the overhead of duplicating one batch of results in
        #   memory is small
        return self.sliced_list[key]

    @Lazy
    def sliced_list(self):
        # We use Lazy here to avoid self.__iter__ giving us new objects every
        # time; in certain cases (such as when database access is necessary)
        # this can be expensive.
        #
        if self.range_memo is None:
            # Legacyy mode - use the slice protocol on results.
            sliced = self.range_factory.getSliceByIndex(self.start, self.end+1)
        else:
            # The range factory gives us a partition on results starting from
            # the previous actual result.
            if self.range_memo is '' and not self.range_forwards:
                # As a special case, when getting the last batch (range_memo
                # None, range_forwards False) we calculate the number of items
                # we would expect from the size of the collection. This is done
                # because that is what the older code did when range_factory
                # was added.
                size = self.listlength % self.size
                if not size:
                    size = self.size
            else:
                # We get one more item than we need, so that we can detect
                # cases where the underlying list ends on a batch boundary
                # without having to check the total size of the list (an
                # expensive operation).
                size = self.size + 1
            sliced = self.range_factory.getSlice(size, self.range_memo,
                self.range_forwards)
            if not self.range_forwards:
                sliced.reverse()
                if self.range_memo is not '' and len(sliced) <= size:
                    if len(sliced) == size:
                        # sliced is meant to have an extra element *after* the
                        # content for end-of-collection-detection, but the
                        # partitioner has just walked backwards, so shuffle things
                        # around to suit. If we didn't get as much data as we asked
                        # for we are at the beginning of the collection and all
                        # elements are needed.
                        sliced = sliced[1:] + sliced[:1]
                    else:
                        # If we got fewer than expected walking backwards, the
                        # range memo may have constrained us. So we need to get
                        # some more results:
                        needed = size - len(sliced)
                        partial = _PartialBatch(sliced)
                        extra_memo = (
                            self.range_factory.getEndpointMemos(partial))
                        extra = self.range_factory.getSlice(needed,
                            extra_memo[1], forwards=True)
                        sliced = sliced + extra
                        self.is_first_batch = True
        # This is the first time we get an inkling of (approximately)
        # how many items are in the list. This is the appropriate time
        # to handle edge cases.
        self._sliced_list = sliced
        return sliced

    def __iter__(self):
        sliced = self.sliced_list
        if len(sliced) > self.size:
            # The slice contains more than a full batch, indicating
            # that there is another batch beyond this one. But we
            # don't actually want to return the item from the next
            # batch.
            sliced = sliced[:-1]
        else:
            # The slice contains a full batch or less, indicating that
            # there is no batch beyond this one. In this case we want
            # to return the full slice.
            pass
        return iter(sliced)

    def first(self):
        return self[0]

    def last(self):
        return self[self.trueSize-1]

    def __contains__(self, item):
        return item in self.__iter__()

    @property
    def has_next_batch(self):
        """See `IBatch`."""
        # self.sliced_list tries to return one more object than the
        # batch size. If it returns the batch size, or fewer, then
        # this batch encompasses the end of the list, for forward
        # batching.
        # In the case of backward batching, an empty memo value
        # indicates that this is the last batch.
        if self.range_forwards:
            return len(self.sliced_list) > self.size
        else:
            return self.range_memo != ''

    def nextBatch(self):
        if not self.has_next_batch:
            return None

        start = self.start + self.size
        memos = self.range_factory.getEndpointMemos(self)
        return _Batch(self.list, self.range_factory, start, self.size,
            range_memo=memos[1], range_forwards=True,
            _listlength=self._listlength)

    @property
    def has_previous_batch(self):
        """See `IBatch`."""
        if self.range_memo is None:
            # If no range memo is specified, sliced_list() falls back
            # to slicing by index. This happens for old URLs, for example.
            return self.start > 0
        if self.range_forwards:
            # For forward batching, we are at the first batch if the memo
            # value is empty.
            return self.range_memo != ''
        else:
            # For backwards batching, we can rely on the flag set in
            # sliced_list. sliced_list is a @Lazy property, so make
            # sure that it has been evaluated.
            self.sliced_list
            return not self.is_first_batch

    def prevBatch(self):
        if not self.has_previous_batch:
            return None

        # The only case in which we should /not/ offer a previous batch
        # is when we are already at position zero, which also happens
        # when the list is empty.
        if self.start <= 0 and self.range_memo is None:
            return None
        start = self.start - self.size
        if start < 0:
            # This situation happens, for instance, when you have a
            # 20-item batch and you manually set your start to 15;
            # in this case, hopping back one batch would be starting at
            # -5, which doesn't really make sense.
            start = 0
        memos = self.range_factory.getEndpointMemos(self)
        return _Batch(self.list, self.range_factory, start, self.size,
            range_memo=memos[0], range_forwards=False,
            _listlength=self._listlength)

    def firstBatch(self):
        return _Batch(self.list, self.range_factory, 0, size=self.size,
            range_memo='', range_forwards=True, _listlength=self._listlength)

    def lastBatch(self):
        # Return the last possible batch for this dataset, at the
        # correct offset.
        last_index = self.listlength - 1
        last_batch_start = last_index - (last_index % self.size)
        if last_batch_start < 0:
            last_batch_start = 0
        return _Batch(self.list, self.range_factory, last_batch_start,
            size=self.size, range_memo='', range_forwards=False,
            _listlength=self._listlength)

    def total(self):
        return self.listlength

    def startNumber(self):
        return self.start+1

    def endNumber(self):
        # If this batch is completely empty, it makes no sense to ask for an
        # "endNumber" so return None.
        if not len(self.sliced_list):
            return None
        return self.start + min(self.size, len(self.sliced_list))



# Copyright 2011 Canonical Ltd.  All rights reserved.
#
# This file is part of lazr.batchnavigator
#
# lazr.batchnavigator is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# lazr.batchnavigator is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with lazr.batchnavigator. If not, see <http://www.gnu.org/licenses/>.

"""Unit tests for lazr.batchnavigator.BatchNavigator."""

import testtools
from testtools.matchers import Equals
from zope.interface.verify import verifyObject
from zope.publisher.browser import TestRequest

from lazr.batchnavigator import BatchNavigator, ListRangeFactory
from lazr.batchnavigator.interfaces import (
    IRangeFactory,
    InvalidBatchSizeError,
    )
from lazr.batchnavigator.z3batching.batch import _Batch
from lazr.batchnavigator.tests.test_z3batching import RecordingFactory

SERVER_URL = 'http://www.example.com/foo'


def query_string(start=None, batch=None, memo=None, direction=None):
    query_string_args = {}
    params = locals()
    def maybe(arg):
        if params[arg] is not None:
            query_string_args[arg] = params[arg]
    maybe('batch')
    maybe('direction')
    maybe('start')
    maybe('memo')
    query_string = "&".join(["%s=%s" % item for item
        in sorted(query_string_args.items())])
    return query_string


def query_url(start=None, batch=None, memo=None, direction=None):
    return SERVER_URL + '?' + query_string(start=start, batch=batch,
        memo=memo,direction=direction)


def batchnav_request(start=None, batch=None, memo=None, direction=None):
    return TestRequest(SERVER_URL=SERVER_URL,
        method='GET', environ={'QUERY_STRING': query_string(start=start,
        batch=batch, memo=memo,direction=direction)})


def sample_batchnav(start=None, batch=None, memo=None, direction=None):
    collection = list(range(1, 11))
    request = batchnav_request(start=start, batch=batch, memo=memo,
        direction=direction)
    range_factory = RecordingFactory(collection)
    return BatchNavigator(collection, request, range_factory=range_factory)


class PickyListLikeCollection:
    """Collection that really hates slices with negative start values.

    Some database-backed collections, e.g. Postgres via Storm, throw
    exceptions if a slice is used that starts with a negative value.  When
    batch navigator is going backwards it is easy to pass such a slice.  This
    collection is used to ensure it is noticed if that happens.
    """
    def __init__(self, collection):
        self._collection = collection
    def __iter__(self):
        return iter(self._collection)
    def __getitem__(self, index):
        if (isinstance(index, slice) and
            index.start is not None and index.start < 0):
            raise RuntimeError
        return self._collection.__getitem__(index)
    def __len__(self):
        return self._collection.__len__()


def empty_batchnav(start=None, batch=None, memo=None, direction=None):
    collection = PickyListLikeCollection([])
    request = batchnav_request(start=start, batch=batch, memo=memo,
        direction=direction)
    range_factory = RecordingFactory(collection)
    return BatchNavigator(collection, request, range_factory=range_factory)


class EqualsQuery(Equals):

    def __init__(self, start=None, batch=None, direction=None, memo=None):
        Equals.__init__(self, query_url(start=start, batch=batch,
            direction=direction, memo=memo))


class TestBatchFactory(testtools.TestCase):

    def test_accepts_range_factory(self):
        # A BatchNavigator can be instantiated with a RangeFactory.
        results = [1, 2, 3, 4]
        range_factory = ListRangeFactory(results)
        request = batchnav_request()
        batchnav = BatchNavigator(
            results, request, range_factory=range_factory)
        self.assertEqual(batchnav.batch.range_factory, range_factory)

    def test_without_memo_direction_gets_non_factory_batch(self):
        # memo and direction were added after the core was developed; for
        # compatability with bookmarks when they are missing the old behaviour
        # is invoked even when an explicit range factory is supplied
        batchnav = sample_batchnav(start=3, batch=3)
        self.assertEqual(list(range(4, 7)), list(batchnav.currentBatch()))
        self.assertEqual([], batchnav.batch.range_factory.calls)

    def test_without_memo_with_backwards_direction_gets_reverse_batch(self):
        # When direction is specified backwards without memo the batch seeks
        # backwards rather than doing a big list protocol slice.
        batchnav = sample_batchnav(start=3, batch=3, direction='backwards')
        self.assertEqual(list(range(10, 11)), list(batchnav.currentBatch()))
        self.assertEqual([('getSlice', 1, '', False)],
            batchnav.batch.range_factory.calls)

    def test_memo_used_for_middle_batch_forwards(self):
        # If we have a memo for a batch, it is used.
        batchnav = sample_batchnav(start=3, batch=3, memo='3')
        self.assertEqual(list(range(4, 7)), list(batchnav.currentBatch()))
        self.assertEqual([('getSlice', 4, '3', True)],
            batchnav.batch.range_factory.calls)

    def test_memo_used_for_middle_batch_backwards(self):
        # If we have a memo for a batch, it is used.
        batchnav = sample_batchnav(start=3, batch=3, memo='6',
            direction='backwards')
        self.assertEqual(list(range(4, 7)), list(batchnav.currentBatch()))
        self.assertEqual([('getSlice', 4, '6', False)],
            batchnav.batch.range_factory.calls)

    def test_lastBatchURL_sets_direction(self):
        batchnav = sample_batchnav(batch=3)
        self.assertThat(batchnav.lastBatchURL(),
            EqualsQuery(start=9, batch=3, direction='backwards'))

    def test_lastBatchURL_not_empty_for_bogus_start_value(self):
        # lastBatchURL() is correct even when the start parameter
        # is bogus: memo says that the batch begins at offset
        # 6, while start points to the last element of the
        # of the batch.
        batchnav = sample_batchnav(start=9, batch=3, memo="6")
        self.assertThat(batchnav.lastBatchURL(),
            EqualsQuery(start=9, batch=3, direction='backwards'))

    def test_firstBatchURL_is_trivial(self):
        batchnav = sample_batchnav(start=3, batch=3, memo='3')
        self.assertThat(batchnav.firstBatchURL(), EqualsQuery(batch=3))

    def test_firstBatchURL_does_not_depend_on_start_parameter(self):
        # nextBatchURL() is correct even when start has the (incorrect)
        # value 0. Start = 0 implies that we are at the first batch.
        # But this value may be wrong (see _Batch.__init__()), and
        # firstBatchURL() does not rely on it.
        batchnav = sample_batchnav(start=0, batch=3, memo='3')
        self.assertThat(batchnav.firstBatchURL(), EqualsQuery(batch=3))

    def test_nextBatchURL_has_memo(self):
        batchnav = sample_batchnav(start=3, batch=3, memo='3')
        self.assertThat(batchnav.nextBatchURL(),
            EqualsQuery(start=6, batch=3, memo='6'))

    def test_prevBatchURL_has_memo_direction(self):
        batchnav = sample_batchnav(start=6, batch=3, memo='6')
        self.assertThat(batchnav.prevBatchURL(),
            EqualsQuery(start=3, batch=3, memo='6', direction='backwards'))

    def test_variable_prefix_affects_all_params(self):
        # subclassing and setting variable_name_prefix causes all the variable
        # names to be adjusted.
        collection = list(range(1, 11))
        request = TestRequest(SERVER_URL=SERVER_URL,
            method='GET', environ={
            'QUERY_STRING': "prefix_start=6&prefix_batch=3&prefix_memo=6"})
        range_factory = RecordingFactory(collection)
        class PrefixBatchNavigator(BatchNavigator):
            variable_name_prefix = 'prefix'
        batchnav = PrefixBatchNavigator(collection, request,
                range_factory=range_factory)
        self.assertThat(batchnav.prevBatchURL(),
            Equals(SERVER_URL + '?prefix_batch=3&prefix_direction=backwards'
                '&prefix_memo=6&prefix_start=3'))

    def test_range_factory_producing_url_unsafe_memos(self):
        # Memo values containing characters with special meaning in URLs
        # are properly escaped by generateBatchURL().
        class WeirdRangeFactory(ListRangeFactory):
            def getEndpointMemos(self, batch):
                return ('start&/1', 'end&/2')

        collection = list(range(1, 11))
        request = TestRequest(
            SERVER_URL=SERVER_URL, method='GET',
            environ={'QUERY_STRING': "start=6&batch=3&memo=6"})
        range_factory = WeirdRangeFactory(collection)
        batchnav = BatchNavigator(
            collection, request, range_factory=range_factory)
        self.assertThat(
            batchnav.nextBatchURL(),
            Equals(SERVER_URL + '?batch=3&memo=end%26%2F2&start=9'))

    def test_empty_collection(self):
        batchnav = empty_batchnav(
            start=2, batch=2)
        self.assertEqual([], list(batchnav.currentBatch()))

    def test_empty_collection_backwards(self):
        batchnav = empty_batchnav(
            start=2, batch=2, direction='backwards')
        self.assertEqual([], list(batchnav.currentBatch()))


class TestListRangeFactory(testtools.TestCase):

    def test_valid_object(self):
        verifyObject(IRangeFactory, ListRangeFactory([]))

    def test_getEndpointMemos_empty(self):
        batch = _Batch([], ListRangeFactory([]))
        self.assertEqual(('0', '0'),
            batch.range_factory.getEndpointMemos(batch))

    def test_getEndpointMemos_one_item(self):
        results = [0, 1, 2]
        batch = _Batch(results, ListRangeFactory(results), start=1, size=1)
        self.assertEqual(('1', '2'),
            batch.range_factory.getEndpointMemos(batch))

    def test_getEndpointMemos_two_items(self):
        results = [0, 1, 2, 3]
        batch = _Batch(results, ListRangeFactory(results), start=1, size=2)
        self.assertEqual(('1', '3'),
            batch.range_factory.getEndpointMemos(batch))

    def test_getEndpointMemos_three_items(self):
        results = [0, 1, 2, 3, 4]
        batch = _Batch(results, ListRangeFactory(results), start=1, size=3)
        self.assertEqual(('1', '4'),
            batch.range_factory.getEndpointMemos(batch))

    def test_getEndpointMemos_start_border(self):
        results = [0, 1]
        batch = _Batch(results, ListRangeFactory(results), size=1)
        self.assertEqual(('0', '1'),
            batch.range_factory.getEndpointMemos(batch))

    def test_getEndpointMemos_end_border(self):
        results = [0, 1]
        batch = _Batch(results, ListRangeFactory(results), start=1, size=1)
        self.assertEqual(('1', '2'),
            batch.range_factory.getEndpointMemos(batch))

    def test_getEndpointMemos_last_batch_end_border(self):
        results = [0, 1, 2, 3]
        batch = _Batch(results, ListRangeFactory(results), start=4, size=1)
        self.assertEqual(('4', '4'),
            batch.range_factory.getEndpointMemos(batch))

    def test_getslice_bad_memo(self):
        self.assertRaises(InvalidBatchSizeError,
            ListRangeFactory([]).getSlice, 3, 'foo')

    def test_getslice_next_end(self):
        # at the end, crickets...
        results = [0, 1, 2, 3, 4, 5]
        _slice = ListRangeFactory(results).getSlice(3, '6')
        self.assertEqual([], _slice)

    def test_getslice_before_start(self):
        # at the end, crickets...
        results = [0, 1, 2, 3, 4, 5]
        _slice = list(ListRangeFactory(results).getSlice(3, '0', forwards=False))
        self.assertEqual([], _slice)

    def test_getslice_empty_slice(self):
        class fakeResults(object):
            def __getitem__(self, item):
                raise Exception('Thou shalt not slice.')
        results = fakeResults()
        ListRangeFactory(results).getSlice(3, '0', forwards=False) # No exception

    def test_getslice_before_end(self):
        results = [0, 1, 2, 3, 4, 5]
        _slice = list(ListRangeFactory(results).getSlice(3, '6', forwards=False))
        self.assertEqual([5, 4, 3], _slice)

    def test_getslice_next(self):
        # The slice returned starts where indicated but continues on.
        results = [0, 1, 2, 3, 4, 5]
        _slice = ListRangeFactory(results).getSlice(3, '3')
        self.assertEqual([3, 4, 5], _slice)

    def test_getslice_before_middle(self):
        # Going backwards does not include the anchor (because going forwards
        # includes it)
        results = [0, 1, 2, 3, 4, 5]
        _slice = list(ListRangeFactory(results).getSlice(3, '3', forwards=False))
        self.assertEqual([2, 1, 0], _slice)

    def test_getSliceByIndex(self):
        # getSliceByIndex returns the slice of the result limited by
        # the indices start, end.
        results = [0, 1, 2, 3, 4, 5]
        _slice = ListRangeFactory(results).getSliceByIndex(2, 5)
        self.assertEqual([2, 3, 4], _slice)

    def test_getSliceByIndex__no_result_set(self):
        # If no result set is present, getSliceByIndex() returns an
        # empty list.
        _slice = ListRangeFactory(None).getSliceByIndex(2, 5)
        self.assertEqual([], _slice)

    def test_picky_collection_ok(self):
        p = PickyListLikeCollection(list(range(5)))
        self.assertEqual(list(range(5)[0:2]), p[0:2])

    def test_picky_collection_bad(self):
        p = PickyListLikeCollection([])
        # It would be nice to demonstrate p[-10:2] but it cannot be done
        # directly since assertRaises needs individual parameters.  The
        # following is the equivalent.
        self.assertRaises(RuntimeError,
                          p.__getitem__, slice(-10, 2, None))

    def test_getSlice_empty_result_set_forwards(self):
        results = PickyListLikeCollection([])
        _slice = ListRangeFactory(results).getSlice(5, forwards=True)
        self.assertEqual([], _slice)

    def test_getSlice_empty_result_set_backwards(self):
        results = PickyListLikeCollection([])
        _slice = ListRangeFactory(results).getSlice(5, forwards=False)
        self.assertEqual([], _slice)

    def test_getSliceByIndex_empty_result_set(self):
        results = PickyListLikeCollection([])
        self.assertRaises(
            RuntimeError,
            ListRangeFactory(results).getSliceByIndex, -1, 1)

    def test_rough_length(self):
        # ListRangeFactory.rough_length returns the length of the
        # result set.
        factory = ListRangeFactory(list(range(4)))
        self.assertEqual(4, factory.rough_length)



# Copyright 2004-2009 Canonical Ltd.  All rights reserved.
#
# This file is part of lazr.batchnavigator
#
# lazr.batchnavigator is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# lazr.batchnavigator is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with lazr.batchnavigator. If not, see <http://www.gnu.org/licenses/>.

from zope.interface import Interface, Attribute

__metaclass__ = type


class InvalidBatchSizeError(AssertionError):
    """Received a batch parameter that was invalid (e.g. exceeds our configured max size)."""


class IRangeFactory(Interface):
    """A helper used to construct better backend queries for batches.

    IRangeFactory instances are paired with a result object that the
    BatchNavigator was constructed with. Generally they will be curried
    as part of determining what backend query to run for the thing being
    batched.
    """

    def getEndpointMemos(batch):
        """Converts a batch to a pair of (lower, upper) endpoint memos.

        :return: A tuple of bytestrings suitable for inclusion in URLs.
        """

    def getSlice(size, endpoint_memo='', forwards=True):
        """Returns a slice of the results starting after endpoint_memo.

        Note that endpoint_memo should be a simple string and is untrusted.

        :param size: The maximum number of entries to return in the slice.
        :param endpoint_memo: An endpoint memo as returned by getEndpointMemos.
            If None or '', the edge of the results is used.
        :param forwards: If True, slice forwards from endpoint_memo. Otherwise
            slice backwards from endpoint_memo (e.g. ascending indices in the
            result will indicate earlier items in the results collection).
        :return: An object honouring the tuple protocol, just like 'results' in
            the IBatchNavigatorFactory constructor call. If forwards was not
            True, the order will be reversed vs the 'results' object, otherwise
            it is identical. The object should be fully materialized from any
            backing store and have no more than size elements in it.
        """

    def getSliceByIndex(start, end):
        """Return a slice of the results.

        :param start: The index of the first element contained in the
            result.
        :param end: The index of the first element not contained in the
            result.
        """

    rough_length = Attribute(
        """The total length of the result set.

        The value does not have to be accurate; it should only be
        used for informational purposes.
        """)


class IBatchNavigator(Interface):
    """A batch navigator for a specified set of results."""

    batch = Attribute("The IBatch for which navigation links are provided.")

    heading = Attribute(
        "The heading describing the kind of objects in the batch.")

    def setHeadings(singular, plural):
        """Set the heading for singular and plural results."""

    def prevBatchURL():
        """Return a URL to the previous chunk of results."""

    def nextBatchURL():
        """Return a URL to the next chunk of results."""


class IBatchNavigatorFactory(Interface):
    def __call__(results, request, start=0, size=None, callback=None,
                 transient_parameters=None, force_start=False,
                 range_factory=None):
        """Constructs an `IBatchNavigator` instance.

        :param results: is an iterable of results.

        :param request: Expected to confirm to
            zope.publisher.interfaces.IRequest.
            The following variables are pulled out of the request to control
            the batch navigator:
              - memo: A record of the underlying storage index pointer for the
                position of the batch.
              - direction: Indicates whether the memo is at the start or end of
                the batch.
              - start: Cosmetic - used to calculate the apparent location (but
                note that due to the concurrent nature of repeated visits to
                batches that the true offset may differ - however the
                collection won't skip or show items twice. For compatibility
                with saved URLs, if memo and direction are both missing then
                start is used to do list slicing into the collection.
              - batch: Controls the amount of items we are showing per batch.
                It will only appear if it's different from the default value
                set when the batch is created.

        :param size: is the default batch size, to fall back to if the
            request does not specify one.  If no size can be determined
            from arguments or request, the launchpad.default_batch_size
            config option is used.

        :param callback: is called, if defined, at the end of object
            construction with the defined batch as determined by the
            start and request parameters.

        :param transient_parameters: optional sequence of parameter
            names that should not be passed on in links generated by
            the batch navigator.  Use this for parameters that had
            meaning when this page was requested, but should not be kept
            around for other page requests in the batch.

        :param force_start: if True, the given `start` argument has
            precedence over the start value in the request, if any.

        :param range_factory: An IRangeFactory paired with results, or None.
            If None, a default IRangeFactory is constructed which simply adapts
            results.

        :raises InvalidBatchSizeError: if the requested batch size is higher
            than the maximum allowed.
        """



##############################################################################
#
# Copyright (c) 2003 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Batch tests.
"""
import operator
import unittest

from zope.interface.verify import verifyObject, verifyClass

from lazr.batchnavigator.z3batching.batch import (
    _Batch,
    BATCH_SIZE,
    _PartialBatch,
    )
from lazr.batchnavigator.z3batching.interfaces import IBatch
from lazr.batchnavigator import ListRangeFactory


class ListWithExplosiveLen(list):
    """A list subclass that doesn't like its len() being called."""
    def __len__(self):
        raise RuntimeError


class ListAllergicToNegativeIndexes(list):
    """A list subclass that doesn't like negative indexes."""
    def __getitem__(self, key):
        if isinstance(key, slice):
            if key.start < 0 or key.stop < 0:
                raise RuntimeError('Negative indexes are not allowed.')
        elif isinstance(key, int):
            if key < 0:
                raise RuntimeError('Negative indexes are not allowed.')
        return super(ListAllergicToNegativeIndexes, self).__getitem__(key)


class ListWithIncorrectLength(list):
    """A list subclass returning a too small length value.

    The goal of configurable range factories is to avoid SQL queries
    involving very expensive counting of result rows.

    One way to avoid a SELECT COUNT(*) FROM ... is to use the number
    of results reported by EXPLAIN SELECT ... FROM ...

    These results are not necessarily precise, hence class _Batch
    must not rely on the length reported by a sequence. This class
    can be used in tests to ensure that result sets with bad length
    information are processed properly.
    """
    def __init__(self, length, data=[]):
        super(ListWithIncorrectLength, self).__init__(data)
        self._length = length

    def __len__(self):
        return self._length

    def __getitem__(self, key):
        # Python 3 needs a little extra help to behave in the odd way we
        # want for tests using this class.
        if isinstance(key, slice):
            if key.start < 0:
                key = slice(max(0, len(self) + key.start), key.stop, key.step)
            if key.stop < 0:
                key = slice(key.start, len(self) + key.stop, key.step)
        return super(ListWithIncorrectLength, self).__getitem__(key)


class RangeFactoryWithValueBasedEndpointMemos:
    """A RangeFactory which uses data values from a batch as endpoint memos.
    """
    def __init__(self, results):
        self.results = results

    def getEndpointMemos(self, batch):
        """See `IRangeFactory`."""
        return batch[0], batch[-1]

    getEndpointMemosFromSlice = getEndpointMemos

    def getSlice(self, size, endpoint_memo='', forwards=True):
        """See `IRangeFactory`."""
        if size == 0:
            return []
        if endpoint_memo == '':
            if forwards:
                return self.results[:size]
            else:
                sliced = self.results[-size:]
                sliced.reverse()
                return sliced

        if forwards:
            index = 0
            while (index < len(self.results) and
                   endpoint_memo >= self.results[index]):
                index += 1
            return self.results[index:index+size]
        else:
            index = len(self.results) - 1
            while (index >= 0 and endpoint_memo < self.results[index]):
                index -= 1
            if index < 0:
                return []
            start_index = max(0, index - size)
            sliced = self.results[start_index:index]
            sliced.reverse()
            return sliced

    def getSliceByIndex(self, start, end):
        """See `IRangeFactory`."""
        return self.results[start:end]

    @property
    def rough_length(self):
        """See `IRangeFactory`."""
        return len(self.results)


class RangeFactoryRecordingResultLengthCalls:
    """A RangeFactory surrogate which records each access to rough_length."""

    def __init__(self):
        self.rough_length_accesses = 0

    @property
    def rough_length(self):
        self.rough_length_accesses += 1
        return 42


class TestingInfrastructureTest(unittest.TestCase):
    def test_ListWithExplosiveLen(self):
        # For some of the tests we want to be sure len() of the underlying
        # collection is never called, so we've created a subclass of list that
        # raises an exception if asked for its length.
        self.assertRaises(RuntimeError, len, ListWithExplosiveLen([1,2,3]))

    def test_ListAllergicToNegativeIndexes(self):
        # Some of the tests want to show that the underlying collection is
        # never accessed with a negative index, so we have a subclass of list
        # that raises an exception if accessed in that way.
        self.assertRaises(
            RuntimeError,
            operator.getitem, ListAllergicToNegativeIndexes([1,2,3]), -1)

    def test_ListWithIncorrectLength(self):
        # Calling len(ListWithIncorrectLength) returns a bogus value.
        self.assertEqual(0, len(ListWithIncorrectLength(0, [1,])))
        self.assertEqual(2, len(ListWithIncorrectLength(2, [1,])))
        # Slicing works for positive indexes, even if len() returns a too
        # small value.
        weird_list = ListWithIncorrectLength(2, [0, 1, 2, 3])
        self.assertEqual([2, 3], weird_list[2:4])
        # But it returns odd data with negative indexes. (This does not
        # matter much because even ListRangeFactory does not use
        # negative indexes.)
        self.assertEqual([0], weird_list[-3:-1])
        # Slicing with positive indexes works for a too large len() value.
        weird_list = ListWithIncorrectLength(6, [0, 1, 2, 3])
        self.assertEqual([2, 3], weird_list[2:4])
        # Slicing with negative indexes returns wrong results,
        self.assertEqual([], weird_list[-2:-1])
        self.assertEqual([3], weird_list[-3:-2])

    def test_RangeFactoryWithValueBased_getEndpointMemos(self):
        data = [str(value) for value in range(10)]
        self.assertEqual(
            ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9'], data)
        factory = RangeFactoryWithValueBasedEndpointMemos(data)
        # The endpoint memo values are the values of the first and last
        # elelemnt of a batch.
        batch = data[:3]
        self.assertEqual(('0', '2'), factory.getEndpointMemos(batch))
        batch = data[4:8]
        self.assertEqual(('4', '7'), factory.getEndpointMemos(batch))
        # getSlice() called with an empty memo value returns the
        # first elements if forwards is True...
        self.assertEqual(
            ['0', '1', '2'],
            factory.getSlice(size=3, endpoint_memo='', forwards=True))
        # ...and the last elements if forwards is False.
        self.assertEqual(
            ['9', '8', '7'],
            factory.getSlice(size=3, endpoint_memo='', forwards=False))
        # A forwards slice starts with a value larger than the
        # given memo value.
        self.assertEqual(
            ['6', '7'],
            factory.getSlice(size=2, endpoint_memo='5', forwards=True))
        # A backwards slice starts with a value smaller than the
        # given memo value.
        self.assertEqual(
            ['4', '3'],
            factory.getSlice(size=2, endpoint_memo='5', forwards=False))
        # A slice is smaller than requested if the end of the results
        # is reached.
        self.assertEqual(
            ['8', '9'],
            factory.getSlice(size=3, endpoint_memo='7', forwards=True))
        self.assertEqual(
            [], factory.getSlice(size=3, endpoint_memo='A', forwards=True))
        self.assertEqual(
            [], factory.getSlice(size=3, endpoint_memo=' ', forwards=False))


class RecordingFactory(ListRangeFactory):
    def __init__(self, results):
        ListRangeFactory.__init__(self, results)
        self.calls = []

    def getEndpointMemos(self, batch):
        self.calls.append('getEndpointMemos')
        return ListRangeFactory.getEndpointMemos(self, batch)

    def getSlice(self, size, memo, forwards):
        self.calls.append(('getSlice', size, memo, forwards))
        return ListRangeFactory.getSlice(self, size, memo, forwards)


class BatchTest(unittest.TestCase):

    def getData(self):
        return ['one', 'two', 'three', 'four', 'five', 'six',
                'seven', 'eight', 'nine', 'ten']

    def test_Interface(self):
        self.assertTrue(IBatch.providedBy(self.getBatch()))

    def getBatch(self, start=0, size=None, range_forwards=None,
                 range_memo=None):
        return _Batch(self.getData(), ListRangeFactory(self.getData()), start,
            size, range_forwards, range_memo)

    def getRecordingBatch(self, start=0, size=None, memo=None, forwards=True):
        range_factory = RecordingFactory(self.getData())
        batch = _Batch(range_factory.results, range_factory, start=start,
            size=size, range_forwards=forwards, range_memo=memo)
        return batch

    def test_constructor(self):
        self.getBatch(9, 3)
        # A start that's larger than the length should still construct a
        # working (but empty) batch.
        self.getBatch(99, 3)
        # If no size is provided a default is used.
        _Batch(self.getData(), ListRangeFactory(self.getData()))

    def test_default_batch_size(self):
        # If we don't specify a batch size, a default one is chosen for us.
        batch = _Batch(self.getData(), ListRangeFactory(self.getData()))
        self.assertEqual(batch.size, BATCH_SIZE)

    def test__len__(self):
        batch = self.getBatch(0, 3)
        self.assertEqual(len(batch), 3)

        batch = self.getBatch(9, 3)
        self.assertEqual(len(batch), 1)

        batch = self.getBatch(99, 3)
        self.assertEqual(len(batch), 0)

        # If the entire set of contents fits into a single batch, the result
        # of len() is correct.
        batch = self.getBatch(size=999)
        self.assertEqual(len(batch), 10)

        # If the entire set of contents fits into a single batch, and we've
        # iterated over the items, the result of len() is correct.
        batch = self.getBatch(size=999)
        list(batch)
        self.assertEqual(len(batch), 10)

    def test__getitem__(self):
        batch = self.getBatch(0, 3)
        self.assertEqual(batch[0], 'one')
        self.assertEqual(batch[1], 'two')
        self.assertEqual(batch[2], 'three')
        batch = self.getBatch(3, 3)
        self.assertEqual(batch[0], 'four')
        self.assertEqual(batch[1], 'five')
        self.assertEqual(batch[2], 'six')
        batch = self.getBatch(9, 3)
        self.assertRaises(IndexError, batch.__getitem__, 3)

        # If the batch is "off the end" and we've materialized the underlying
        # list (by iterating over the batch) we will get an IndexError if we
        # pass in a negative item index.
        batch = self.getBatch(99, 3)
        list(iter(batch))
        self.assertRaises(IndexError, batch.__getitem__, -1)

    def test__iter__(self):
        batch = self.getBatch(0, 3)
        self.assertEqual(list(iter(batch)), ['one', 'two', 'three'])
        batch = self.getBatch(9, 3)
        self.assertEqual(list(iter(batch)), ['ten'])
        batch = self.getBatch(99, 3)
        self.assertEqual(list(iter(batch)), [])

    def test__getitem__does_not_use_negative_indices(self):
        # Some collections don't implement __len__, so slicing them doesn't
        # work, therefore we want to be sure negative slices are not passed
        # through to the underlying collection.
        data = ListAllergicToNegativeIndexes(self.getData())
        batch = _Batch(data, ListRangeFactory(data), 99, 3)
        list(iter(batch))
        self.assertRaises(IndexError, batch.__getitem__, -1)

    def test__contains__(self):
        batch = self.getBatch(0, 3)
        self.assertTrue(batch.__contains__('one'))
        self.assertTrue(batch.__contains__('two'))
        self.assertTrue(batch.__contains__('three'))
        self.assertFalse(batch.__contains__('four'))
        batch = self.getBatch(6, 3)
        self.assertFalse(batch.__contains__('one'))
        self.assertTrue(batch.__contains__('seven'))
        self.assertFalse(batch.__contains__('ten'))

    def test_last_batch_len_optimization(self):
        # If the current batch is known to be the last batch, then we can
        # calculate the total number of items without calling len() on the
        # underlying collection.
        data = ListWithExplosiveLen(self.getData())
        batch = _Batch(data, ListRangeFactory(data), 0, 20)

        # If we get the total number of items before accessing the underlying
        # data, we'll call len() on the collection object.
        self.assertRaises(RuntimeError, lambda: batch.listlength)

        # If instead we do something that materialized the slice of the
        # underlying collection that this batch represents...
        next(iter(batch))
        # ... then no exception is raised and we get the number of items.
        self.assertEqual(batch.listlength, 10)

    def test_firstBatch(self):
        """Check that the link to the first batch works.

        This first batch will be always pointing to the first available batch
        and, its main difference with the 'prev' and 'next' batches is, that
        will not be None ever.
        """
        # first batch when we are at the beginning of the batch.
        first = self.getBatch(0, 3).firstBatch()
        self.assertEqual(list(iter(first)), ['one', 'two', 'three'])
        # first batch when we are in the second set of items of the batch.
        first = self.getBatch(3, 3).firstBatch()
        self.assertEqual(list(iter(first)), ['one', 'two', 'three'])
        # first batch when we are in the third set of items of the batch.
        first = self.getBatch(6, 3).firstBatch()
        self.assertEqual(list(iter(first)), ['one', 'two', 'three'])
        # first batch when we are at the end of the batch.
        first = self.getBatch(9, 3).firstBatch()
        self.assertEqual(list(iter(first)), ['one', 'two', 'three'])
        # first batch when we get a request for an out of range item.
        first = self.getBatch(99, 3).firstBatch()
        self.assertEqual(list(iter(first)), ['one', 'two', 'three'])

    def test_nextBatch(self):
        next = self.getBatch(0, 3).nextBatch()
        self.assertEqual(list(iter(next)), ['four', 'five', 'six'])
        nextnext = next.nextBatch()
        self.assertEqual(list(iter(nextnext)), ['seven', 'eight', 'nine'])
        next = self.getBatch(9, 3).nextBatch()
        self.assertEqual(next, None)
        next = self.getBatch(99, 3).nextBatch()
        self.assertEqual(next, None)

    def test_nextBatch__backwards_batching(self):
        # The "first backwards" batch has no next batch.
        batch = self.getBatch(9, 3, range_forwards=False, range_memo='')
        self.assertTrue(batch.nextBatch() is None)
        # previous batches of this batch have a batch.
        batch = batch.prevBatch()
        self.assertFalse(batch.nextBatch() is None)

    def test_prevBatch(self):
        prev = self.getBatch(9, 3).prevBatch()
        self.assertEqual(list(iter(prev)), ['seven', 'eight', 'nine'])
        prevprev = prev.prevBatch()
        self.assertEqual(list(iter(prevprev)), ['four', 'five', 'six'])
        prev = self.getBatch(0, 3).prevBatch()
        self.assertEqual(prev, None)
        prev = self.getBatch(2, 3).prevBatch()
        self.assertEqual(list(iter(prev)), ['one', 'two', 'three'])

        # If we create a batch that's out of range, and don't get its
        # length or any of its data, its previous batch will also be
        # out of range.
        out_of_range = self.getBatch(99, 3)
        out_of_range_2 = out_of_range.prevBatch()
        self.assertEqual(out_of_range_2.start, 96)

    def test_lastBatch(self):
        """Check that the link to the last batch works.

        This last batch will be always pointing to the last available batch
        and, its main difference with the 'prev' and 'next' batches is, that
        will not be None ever.
        """
        # last batch when we are at the beginning of the batch.
        last = self.getBatch(0, 3).lastBatch()
        self.assertEqual(list(iter(last)), ['ten'])
        # last batch when we are in the second set of items of the batch.
        last = self.getBatch(3, 3).lastBatch()
        self.assertEqual(list(iter(last)), ['ten'])
        # last batch when we are in the third set of items of the batch.
        last = self.getBatch(6, 3).lastBatch()
        self.assertEqual(list(iter(last)), ['ten'])
        # last batch when we are at the end of the batch.
        last = self.getBatch(9, 3).lastBatch()
        self.assertEqual(list(iter(last)), ['ten'])
        # last batch when we get a request for an out of range item.
        last = self.getBatch(99, 3).lastBatch()
        self.assertEqual(list(iter(last)), ['ten'])

        # We are going to test now the same, but when we get a request of 5
        # items per batch because we had a bug in the way we calculate the
        # last batch set that was only happening when we were using a batch
        # size that is multiple of the item list length.

        # last batch when we are at the beginning of the batch.
        last = self.getBatch(0, 5).lastBatch()
        self.assertEqual(
            list(iter(last)), ['six', 'seven', 'eight', 'nine', 'ten'])
        # last batch when we are in the second set of items of the batch.
        last = self.getBatch(5, 5).lastBatch()
        self.assertEqual(
            list(iter(last)), ['six', 'seven', 'eight', 'nine', 'ten'])
        # last batch when we get a request for an out of range item.
        last = self.getBatch(99, 5).lastBatch()
        self.assertEqual(
            list(iter(last)), ['six', 'seven', 'eight', 'nine', 'ten'])

    def test_batchRoundTrip(self):
        batch = self.getBatch(0, 3).nextBatch()
        self.assertEqual(list(iter(batch.nextBatch().prevBatch())),
                         list(iter(batch)))

    def test_first_last(self):
        batch = self.getBatch(0, 3)
        self.assertEqual(batch.first(), 'one')
        self.assertEqual(batch.last(), 'three')
        batch = self.getBatch(9, 3)
        self.assertEqual(batch.first(), 'ten')
        self.assertEqual(batch.last(), 'ten')
        batch = self.getBatch(99, 3)
        self.assertRaises(IndexError, batch.first)
        self.assertRaises(IndexError, batch.last)

    def test_total(self):
        batch = self.getBatch(0, 3)
        self.assertEqual(batch.total(), 10)
        batch = self.getBatch(6, 3)
        self.assertEqual(batch.total(), 10)
        batch = self.getBatch(99, 3)
        self.assertEqual(batch.total(), 10)

    def test_trueSize_of_full_batch(self):
        # The .trueSize property reports how many items are in the batch.
        batch = self.getBatch(0, 3)
        self.assertEqual(batch.trueSize, 3)

    def test_trueSize_of_last_batch(self):
        # If the current batch contains fewer than a full batch, .trueSize
        # reports how many items are in the batch.
        batch = self.getBatch(9, 3)
        self.assertEqual(batch.trueSize, 1)

    def test_optimized_trueSize_of_full_batch(self):
        # If the unerlying items in the batch have been materialized, there is
        # a code path that avoids calling len() on the underlying collection.
        data = ListWithExplosiveLen(self.getData())
        batch = _Batch(data, ListRangeFactory(data), 0, 3)
        next(iter(batch)) # Materialize the items.
        self.assertEqual(batch.trueSize, 3)

    def test_optimized_trueSize_of_last_batch(self):
        # If the current batch contains fewer than a full batch, .trueSize
        # reports how many items are in the batch.
        data = ListWithExplosiveLen(self.getData())
        batch = _Batch(data, ListRangeFactory(data), 9, 3)
        next(iter(batch)) # Materialize the items.
        self.assertEqual(batch.trueSize, 1)

    def test_startNumber(self):
        batch = self.getBatch(0, 3)
        self.assertEqual(batch.startNumber(), 1)
        batch = self.getBatch(9, 3)
        self.assertEqual(batch.startNumber(), 10)
        batch = self.getBatch(99, 3)
        self.assertEqual(batch.startNumber(), 100)

    def test_endNumber(self):
        # If a batch of size 3 starts at index 0, the human-friendly numbering
        # of the last item should be 3.
        batch = self.getBatch(0, 3)
        self.assertEqual(batch.endNumber(), 3)
        # If a batch of size 3 starts at index 9 and has just 1 item in it
        # (because there are 10 items total), the human-friendly numbering
        # of the last item should be 10.
        batch = self.getBatch(9, 3)
        self.assertEqual(batch.endNumber(), 10)
        # If the batch start exceeds the number number of items, then the
        # endNumber will be None to indicate that it makes no sense to ask.
        batch = self.getBatch(99, 3)
        self.assertEqual(batch.endNumber(), None)

    def test_sliced_list_uses_range_factory(self):
        batch = self.getRecordingBatch(memo='', forwards=False)
        list(batch)
        self.assertEqual([('getSlice', 10, '', False)], batch.range_factory.calls)

    def test_firstBatch_generates_memo_None(self):
        batch = self.getRecordingBatch(start=3, size=3, memo='3')
        first = batch.firstBatch()
        self.assertEqual('', first.range_memo)
        self.assertEqual(True, first.range_forwards)
        self.assertEqual(0, first.start)
        self.assertEqual(3, first.size)
        self.assertEqual(['one', 'two', 'three'], list(first))

    def test_lastBatch_generates_memo_None_backwards(self):
        batch = self.getRecordingBatch(start=3, size=3, memo='3')
        last = batch.lastBatch()
        self.assertEqual('', last.range_memo)
        self.assertEqual(False, last.range_forwards)
        # len() is known, and size() is known, so the last batch starts at
        # len-size
        self.assertEqual(9, last.start)
        self.assertEqual(3, last.size)
        self.assertEqual(['ten'], list(last))

    def test_nextBatch_generates_upper_memo(self):
        batch = self.getRecordingBatch(start=3, size=3, memo='3')
        next = batch.nextBatch()
        self.assertEqual('6', next.range_memo)
        self.assertEqual(True, next.range_forwards)
        self.assertEqual(6, next.start)
        self.assertEqual(3, next.size)
        self.assertEqual(['seven', 'eight', 'nine'], list(next))

    def test_prevBatch_generates_lower_memo_backwards(self):
        batch = self.getRecordingBatch(start=6, size=3, memo='3')
        prev = batch.prevBatch()
        self.assertEqual('6', prev.range_memo)
        self.assertEqual(False, prev.range_forwards)
        # len() is known, and size() is known, so the prev batch starts at
        # len-size
        self.assertEqual(3, prev.start)
        self.assertEqual(3, prev.size)
        self.assertEqual(['four', 'five', 'six'], list(prev))

    def test_backwards_memo_inside_first_batch(self):
        # If data is retrieved backwards from the memo point and if
        # there is less data available than requested, additional
        # data stored "after the endpoint" is returned.
        batch = self.getRecordingBatch(start=0, size=5, memo=3, forwards=False)
        self.assertEqual(['one', 'two', 'three', 'four', 'five'], list(batch))

    def test_trueSize__with_bad_list_length(self):
        # Even if the len(resultset) claims to have only 2 elements,
        # batches starting at offsets 1 and 2 have two elements.
        weird_list = ListWithIncorrectLength(2, [0, 1, 2, 3])
        batch = _Batch(
            weird_list, range_factory=ListRangeFactory(weird_list),
            size=2, start=1)
        self.assertEqual(2, batch.trueSize)
        self.assertEqual([1, 2, 3], batch.sliced_list)
        batch = _Batch(
            weird_list, range_factory=ListRangeFactory(weird_list),
            size=2, start=2)
        self.assertEqual(2, batch.trueSize)
        self.assertEqual([2, 3], batch.sliced_list)

    def test_nextBatch__too_small_results_length(self):
        # Batch.nextBatch() is not None even if len(results)
        # indicates that there should not be a next batch.
        weird_list = ListWithIncorrectLength(3, [0, 1, 2, 3, 4])
        batch = _Batch(
            weird_list, range_factory=ListRangeFactory(weird_list),
            size=3)
        self.assertTrue(batch.nextBatch() is not None)

    def test_nextBatch__bad_results_length(self):
        # Even if len(results) claims that there are only 6 results,
        # batch.nextBatch() is not None
        weird_list = ListWithIncorrectLength(6, [0, 1, 2, 3, 4, 5, 6])
        batch = _Batch(
            weird_list, range_factory=ListRangeFactory(weird_list),
            size=3, range_memo='3', start=3, range_forwards=True)
        self.assertEqual([3, 4, 5, 6], batch.sliced_list)
        self.assertEqual([6], batch.nextBatch().sliced_list)

    def test_prevBatch__outdated_start_params(self):
        # The start parameter of a backwards batch may reach zero
        # even when the real begin of a result set is not yet reached.
        data = [1, 2, 3, 4, 5, 6]
        range_factory = RangeFactoryWithValueBasedEndpointMemos(data)
        # We create the first backwards.
        batch = _Batch(
            data, range_factory=range_factory, size=3, range_memo='',
            start=3, range_forwards=False)
        self.assertEqual([4, 5, 6], batch.sliced_list)
        self.assertEqual(3, batch.start)
        # The previous batch is now the first batch.
        batch = batch.prevBatch()
        self.assertEqual(0, batch.start)
        self.assertEqual(4, batch.range_memo)
        self.assertEqual([1, 2, 3, 4], batch.sliced_list)
        self.assertTrue(batch.prevBatch() is None)
        # If we now insert another value at the start of the result set,
        # and when we reload this batch with the same range parameters,
        # we get the same data slice, but also a prevBtach()
        # that is not None.
        data = [0, 1, 2, 3, 4, 5, 6]
        range_factory = RangeFactoryWithValueBasedEndpointMemos(data)
        new_batch = _Batch(
            data, range_factory=range_factory, size=3,
            range_memo=batch.range_memo, start=batch.start,
            range_forwards=False)
        # the contents of the batch is still correct. (Note that the
        # value of the last does not matter, see _Batch.sliced_list().)
        self.assertEqual([1, 2, 3, 0], new_batch.sliced_list)
        # And we get a previous batch.
        self.assertEqual([0, 1, 2, 3], new_batch.prevBatch().sliced_list)
        # Note that both new_batch and its previous batch claim to start
        # at 0.
        self.assertEqual(0, new_batch.start)
        self.assertEqual(0, new_batch.prevBatch().start)

    def test_last_backwards_batch_with_value_based_range_factory(self):
        # Another slice is added in _Batch.sliced_list() when the
        # regular slice of a backwards batch does not return the
        # number of required elements. This works for range factories
        # which are based on the values too.
        data = [str(value) for value in range(10)]
        range_factory = RangeFactoryWithValueBasedEndpointMemos(data)
        batch = _Batch(
            data, range_factory=range_factory, size=3, range_memo='1',
            start=1, range_forwards=False)
        self.assertEqual(['0', '1', '2', '3'], batch.sliced_list)

    def test_PartialBatch(self):
        # PartialBatch implements the full IBatch interface.
        self.assertTrue(verifyClass(IBatch, _PartialBatch))
        partial = _PartialBatch(sliced_list=list(range(3)))
        self.assertTrue(verifyObject(IBatch, partial))
        # trueSize is the length of sliced_list
        self.assertEqual(3, partial.trueSize)
        # sliced_list is passed by the contrucotr parameter sliced_list
        self.assertEqual([0, 1, 2], partial.sliced_list)
        # __len__() returns the length of the sliced list
        self.assertEqual(3, len(partial))
        # __iter__() iterates over sliced_list
        self.assertEqual([0, 1, 2], [element for element in partial])
        # __contains__() works.
        self.assertTrue(1 in partial)
        self.assertFalse(3 in partial)
        # prevBatch(), nextBatch() exost but are not implemented.
        self.assertRaises(NotImplementedError, partial.prevBatch)
        self.assertRaises(NotImplementedError, partial.nextBatch)
        # first and last are implemented.
        self.assertEqual(0, partial.first())
        self.assertEqual(2, partial.last())
        # total() return the length of sliced_list
        self.assertEqual(3, partial.total())
        # startNumber, endNumber() are implemented
        self.assertEqual(1, partial.startNumber())
        self.assertEqual(4, partial.endNumber())

    def test_has_next_batch__forwards_with_memo_not_at_end(self):
        # If a batch is not the last batch for a result set,
        # has_next_batch is True. When the parameter range_memo
        # is given, the value of start does not matter.
        batch = self.getBatch(
            start=7, size=3, range_forwards=True, range_memo="6")
        self.assertTrue(batch.has_next_batch)

    def test_has_next_batch__forwards_with_memo_at_end(self):
        # If a batch is the last batch for a result set,
        # has_next_batch is False. When the parameter range_memo
        # is given, the value of start does not matter.
        batch = self.getBatch(
            start=6, size=3, range_forwards=True, range_memo="7")
        self.assertFalse(batch.has_next_batch)

    def test_has_next_batch__forwards_without_memo_not_at_end(self):
        # If a batch is the last batch for a result set,
        # has_next_batch is False. When the parameter range_memo
        # is not given, the value of start is used.
        batch = self.getBatch(start=6, size=3)
        self.assertTrue(batch.has_next_batch)

    def test_has_next_batch__forwards_without_memo_at_end(self):
        # If a batch is the last batch for a result set,
        # has_next_batch is False. When the parameter range_memo
        # is not given, the value of start is used.
        batch = self.getBatch(start=7, size=3)
        self.assertFalse(batch.has_next_batch)

    def test_has_previous_batch__backwards_with_memo_not_at_end(self):
        # If a batch is not the first batch for a result set,
        # has_previous_batch is True. The value of start does not
        # matter.
        batch = self.getBatch(
            start=0, size=3, range_forwards=False, range_memo="4")
        self.assertTrue(batch.has_previous_batch)

    def test_has_previous_batch__backwards_with_memo_at_end(self):
        # If a batch is not the first batch for a result set,
        # has_previous_batch is False.The value of start does not
        # matter.
        batch = self.getBatch(
            start=1, size=3, range_forwards=False, range_memo="3")
        self.assertFalse(batch.has_previous_batch)

    def test_batch_gets_listlength_from_factory(self):
        # Batches get the length of the result set from the
        # range factory.
        range_factory = RangeFactoryRecordingResultLengthCalls()
        batch = _Batch(list(range(3)), range_factory)
        listlength = batch.listlength
        self.assertEqual(42, listlength)
        self.assertEqual(1, range_factory.rough_length_accesses)



##############################################################################
#
# Copyright (c) 2003 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.0 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Batching Support

$Id$
"""
from zope.interface import Attribute
from zope.interface.common.mapping import IItemMapping

class IBatch(IItemMapping):
    """A Batch represents a sub-list of the full enumeration.

    The Batch constructor takes a list (or any list-like object) of elements,
    a starting index and the size of the batch. From this information all
    other values are calculated.
    """

    def __len__():
        """Return the length of the batch. This might be different than the
        passed in batch size, since we could be at the end of the list and
        there are not enough elements left to fill the batch completely."""

    def __iter__():
        """Creates an iterator for the contents of the batch (not the entire
        list)."""

    def __getitem__(index):
        """Return the element at the given offset."""

    def __contains__(key):
        """Checks whether the key (in our case an index) exists."""

    def nextBatch():
        """Return the next batch. If there is no next batch, return None."""

    def prevBatch():
        """Return the previous batch. If there is no previous batch, return
        None."""

    def first():
        """Return the first element of the batch."""

    def last():
        """Return the last element of the batch."""

    def total():
        """Return the length of the list (not the batch)."""

    def startNumber():
        """Give the start **number** of the batch, which is 1 more than the
        start index passed in."""

    def endNumber():
        """Give the end **number** of the batch, which is 1 more than the
        final index."""

    sliced_list = Attribute(
        "A sliced list as returned by IRangeFactory.sliced_list.")

    trueSize = Attribute("The actual size of this batch.")

    has_previous_batch = Attribute(
        "True, if this batch has a predecessor, else False.")

    has_next_batch = Attribute(
        "True, if this batch has a successor, else False.")



# Copyright 2009 Canonical Ltd.  All rights reserved.
#
# This file is part of lazr.batchnavigator
#
# lazr.batchnavigator is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# lazr.batchnavigator is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with lazr.batchnavigator. If not, see <http://www.gnu.org/licenses/>.
"The lazr.batchnavigator tests."

import unittest

from lazr.batchnavigator.tests import (
    test_docs,
    )

def test_suite():
    # Discovery would be nice, but needs python2.7 for use with subunit &
    # testrepository.
    result = unittest.TestSuite()
    result.addTests(test_docs.additional_tests())
    modules = ['batchnavigator', 'z3batching']
    result.addTests(unittest.TestLoader().loadTestsFromNames([
        'lazr.batchnavigator.tests.test_' + name for name in modules]))
    return result



#!/usr/bin/env python

# Copyright 2009 Canonical Ltd.  All rights reserved.
#
# This file is part of lazr.batchnavigator
#
# lazr.batchnavigator is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# lazr.batchnavigator is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with lazr.batchnavigator. If not, see <http://www.gnu.org/licenses/>.

from setuptools import setup, find_packages

# generic helpers primarily for the long_description
def generate(*docname_or_string):
    marker = '.. pypi description ends here'
    res = []
    for value in docname_or_string:
        if value.endswith('.rst'):
            with open(value) as f:
                value = f.read()
            idx = value.find(marker)
            if idx >= 0:
                value = value[:idx]
        res.append(value)
        if not value.endswith('\n'):
            res.append('')
    return '\n'.join(res)
# end generic helpers

tests_require = [
    'testtools',
    'zope.publisher',
    'zope.testrunner',
    ]

setup(
    name='lazr.batchnavigator',
    version='1.3.1',
    namespace_packages=['lazr'],
    packages=find_packages('src'),
    package_dir={'': 'src'},
    include_package_data=True,
    zip_safe=False,
    maintainer='LAZR Developers',
    maintainer_email='lazr-users@lists.launchpad.net',
    description=open('README.rst').readline().strip(),
    long_description=generate(
        'src/lazr/batchnavigator/docs/index.rst',
        'NEWS.rst'),
    license='LGPL v3',
    install_requires=[
        'importlib-metadata; python_version < "3.8"',
        'setuptools',
        'six',
        'zope.cachedescriptors',
        'zope.interface>=3.6.0',
        ],
    url='https://launchpad.net/lazr.batchnavigator',
    download_url='https://launchpad.net/lazr.batchnavigator/+download',
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        ],
    tests_require=tests_require,
    extras_require=dict(
        docs=['Sphinx'],
        test=tests_require,
    ),
    test_suite='lazr.batchnavigator.tests',
    )



# Copyright 2009 Canonical Ltd.  All rights reserved.
#
# This file is part of lazr.batchnavigator
#
# lazr.batchnavigator is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# lazr.batchnavigator is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
# License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with lazr.batchnavigator. If not, see <http://www.gnu.org/licenses/>.
"Test harness for doctests."

from __future__ import print_function

__metaclass__ = type
__all__ = [
    'load_tests',
    ]

import atexit
import doctest
import os
import pkg_resources

DOCTEST_FLAGS = (
    doctest.ELLIPSIS |
    doctest.NORMALIZE_WHITESPACE |
    doctest.REPORT_NDIFF)


def test_raises(exc_class, method, *args, **kwargs):
    try:
        method(*args, **kwargs)
    except Exception as e:
        if isinstance(e, exc_class):
            print(e)
            return
        raise
    raise Exception("Expected exception %s not raised" % exc_class)


def load_tests(loader, tests, pattern):
    """Load all the doctests."""
    atexit.register(pkg_resources.cleanup_resources)
    doctest_files = []
    if pkg_resources.resource_exists('lazr.batchnavigator', 'docs'):
        for name in pkg_resources.resource_listdir(
                'lazr.batchnavigator', 'docs'):
            if name.endswith('.rst'):
                doctest_files.append(
                    os.path.abspath(
                        pkg_resources.resource_filename(
                            'lazr.batchnavigator', 'docs/%s' % name)))
    tests.addTest(doctest.DocFileSuite(
        *doctest_files, module_relative=False, optionflags=DOCTEST_FLAGS,
        globs={
            "print_function": print_function,
            "test_raises": test_raises,
            }))
    return tests
